unit BitToTibConverter;

// - VCC -
// Xilinx .bit file to reversed .bit file converter
// For Spartan-3(E) and Spartan-6 FPGAs
// - 2010 -
// - 2014 - made Delphi XE5 compatible

interface

uses
  Windows, SysUtils, Classes;

type
  TXilinxBitFile = class(TComponent)
  private
    FBitLength: Cardinal;   { length of bistream in bytes }
    FBitFileName: TFileName;   { .bit FileName }
    FBitFileCurrentPostion: Cardinal; {Current position in bitfile data section}
    FF: File;

    function BitReadField: AnsiString;

  public
    constructor Create(AOwner: TComponent); override;
    function OpenBitFile: Boolean;
    procedure CloseBitFile;

    function BitReadData(SizeToRead: Cardinal): AnsiString;
    function GetNumberOfDataBits: Cardinal;

    property BitLength: Cardinal read FBitLength; {Length of bistream in bytes}
  end;


  {The result is a .tib file with the same name and path as the .bit file.}
  function ConvertBitFileToTibFile(BitFileName: TFileName; var NumberOfDataBits: Integer): TFileName;

implementation



{Converts a Xilinx .bit file to a reversed .bit file (i.e. the .tib file)}
{The FPGA needs the .tib file. The function returns the path to the .tib file}
function ConvertBitFileToTibFile(BitFileName: TFileName; var NumberOfDataBits: Integer): TFileName;
var
  TibFileName: TFileName;
  FXBit: TXilinxBitFile;
  BitContent: AnsiString;
  TibContent: AnsiString;
  i: Integer;
  GG: File;
begin
  TibFileName := BitFileName;
  Delete(TibFileName, Length(TibFileName) - 3, 4);   //not sure if this test works on XE5
  TibFileName := TibFileName + '.tib';
  Result := TibFileName;

  if not FileExists(BitFileName) then
  begin
    MessageBox(0, PChar(BitFileName + ' does not exist.'), 'Bit to Tib converter', MB_ICONERROR);
    Exit;
  end;

  FXBit := TXilinxBitFile.Create(nil);
  FXbit.FBitFileName := BitFileName;
  try
    if not FXBit.OpenBitFile then
      MessageBox(0, 'Error: The .bit file can''t be parsed.', 'Bit to Tib converter', MB_ICONERROR)
    else
    begin
      FXBit.BitReadData(0); {Get Bit Length}
      NumberOfDataBits := FXBit.BitLength;    {output this info}

      BitContent := FXBit.BitReadData(FXBit.BitLength);    {The computer has enough memory to read the entire file in a single operation but a microctroller needs to call BitReadData for less data.}

      FXBit.CloseBitFile;

      SetLength(TibContent, Length(BitContent));
      for i := 1 to Length(BitContent) do
        TibContent[i] := BitContent[Length(BitContent) - i + 1];

      AssignFile(GG, TibFileName);
      Rewrite(GG, 1);
      BlockWrite(GG, TibContent[1], Length(TibContent));
      CloseFile(GG);
    end;
  finally
    FXBit.Free;
  end;
end;




{Creates the class and generates the lookup table.}
constructor TXilinxBitFile.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FBitFileCurrentPostion := 0;
end;


{Loads the BitFile specified by "BitFileName" property and reads the header.
 Returns True for success. The file is closed only in case of an error.}
function TXilinxBitFile.OpenBitFile: Boolean;
var
  BitFileHeader: array[0..12] of Byte;
  Field: AnsiChar;  {must be 1 byte in size - do not use WideChar here !}
begin
  Result := False;
  Field := #0;

  if not FileExists(FBitFileName) then
    Exit;

  AssignFile(FF, FBitFileName);
  Reset(FF, 1);
  if FileSize(FF) = 0 then
  begin
    CloseFile(FF);
    Exit;
  end;

  try
    BlockRead(FF, BitFileHeader, 13);
    if Eof(FF) then
    begin
      Result := False;
      Exit;
    end;

    while True do
    begin
      if Eof(FF) then
        Exit;

      BlockRead(FF, Field, 1);
      case Field of
        'a', 'b', 'c', 'd': BitReadField;  {different data about FPGA used, timestamp etc}
        'e': Break; {Start of data}
      end;
    end;

  except
    Result := False;
    Exit;
  end;  

  Result := True;  
end;


{Closes the bit file.}
procedure TXilinxBitFile.CloseBitFile;
begin
  try
    CloseFile(FF);
  except
  end;  
end;



{Reads an info field from the bit file header.}
function TXilinxBitFile.BitReadField: AnsiString;
var
  t: array[0..1] of Byte;
  len: Word;
  i: Integer;
  Ch: AnsiChar; {must be 1 byte in size - do not use WideChar here !}
  Field: AnsiString;
begin
  Field := '';
  BlockRead(FF, t, 2);
  len := t[0] shl 8 + t[1];

  for i := 0 to len - 1 do
  begin
    BlockRead(FF, Ch, 1);
    if Ch > #0 then
      field := field + Ch;
  end;
  
  Result := Field;
end;


{Reads "SizeToRead" bytes from bit file data section. Returns data read.
 If "SizeToRead" is greater than the available data, the function returns only
 the available data.}
function TXilinxBitFile.BitReadData(SizeToRead: Cardinal): AnsiString;
var
  t: array[0..3] of Byte;
  b: Byte;
  i: Cardinal;
  Buffer: AnsiString;
begin
  if (FBitFileCurrentPostion = 0) and (SizeToRead = 0) then
  begin
    BlockRead(FF, t, 4);
    FBitLength := t[0] shl 24 + t[1] shl 16 + t[2] shl 8 + t[3];
  end;

  if SizeToRead = 0 then
    Exit;
  
  SetLength(Buffer, SizeToRead);

  for i := 0 to SizeToRead - 1 do
  begin
    if Eof(FF) or (FBitFileCurrentPostion + SizeToRead > FBitLength) then
    begin
      SetLength(Buffer, i);
      Inc(FBitFileCurrentPostion, i);
      Result := Buffer;
      Exit;
    end;
    BlockRead(FF, b, 1);
    Buffer[SizeToRead - i] := AnsiChar(Chr(b));
  end;

  Inc(FBitFileCurrentPostion, SizeToRead);
  Result := Buffer;
end;


{Returns the number of data bits.}
function TXilinxBitFile.GetNumberOfDataBits: Cardinal;
begin
  Result := FBitLength * 8;
end;  


end.
